/*
 * Copyright (c) 2008, Felix Opatz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of  nor the names of its contributors may be used to
 *    endorse or promote products derived from this software without specific
 *    prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * Date        Change  
 * ---------------------------------------------------------------------------
 * 2008-10-28  Initial implementation.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#ifdef WIN32
#include <windows.h>
#endif

#include "zjlink.h"

/* Print information about zjlink */
void PrintHelp(void)
{
    puts("Sends or receives files over serial COM lines (nullmodems).\n");
    puts("zjlink [-p port] [-b baudrate] [-m mode] [file(s)]");
    putchar('\n');
    puts("  -p port        use port for sending (e.g. COM3)");
    puts("  -b baudrate    use baudrate (e.g. 1200 for really bad lines)");
    puts("  -m mode        set mode (e.g. 8N1, 7O1, 6E2)");
    puts("                 if you don't understand this you won't need this");
    putchar('\n');
    puts("Started without filenames will put zjlink into receive mode.");
    printf("Default settings: -p %s -b %lu -m %s.\n", DEF_COMPORT, DEF_BAUDRATE, DEF_MODESTRING);
    putchar('\n');
    puts("For bugreports or comments, please visit (URL:http://www.zotteljedi.de/).");

    putchar('\n');
}

/* Reads a WORD (MSB first) */
WORD ReadWord(HPORT hPort)
{
    WORD wRet = 0;
    int iCounter = sizeof(wRet);

    /* until all bytes are read */
    while (iCounter-- > 0)
    {
        /* shift existing bytes */
        wRet <<= 8;
        /* insert next byte */
        wRet |= ReadByte(hPort);
    }

    return wRet;
}

/* Reads a DWORD (MSB first) */
DWORD ReadDword(HPORT hPort)
{
    DWORD dwRet = 0;
    int iCounter = sizeof(dwRet);

    /* until all bytes are read */
    while (iCounter-- > 0)
    {
        /* shift existing bytes */
        dwRet <<= 8;
        /* insert next byte */
        dwRet |= ReadByte(hPort);
    }

    return dwRet;
}

/* Read a block of wBytesToRead bytes into pbBuffer */
void ReadBlock(HPORT hPort, PBYTE pbBuffer, WORD wBytesToRead)
{
    BYTE bByte;

    /* until all bytes are read */
    while (wBytesToRead > 0)
    {
        /* read next byte */
        bByte = ReadByte(hPort);
        /* save into buffer */
        *pbBuffer = bByte;
        /* prepare for next one */
        pbBuffer++;
        wBytesToRead--;
    }
}

/* Writes a WORD (MSB first) */
void WriteWord(HPORT hPort, WORD wData)
{
    WriteByte(hPort, (wData >> 8) & 0xFF);
    WriteByte(hPort, wData & 0xFF);
}

/* Writes a DWORD (MSB first) */
void WriteDword(HPORT hPort, DWORD dwData)
{
    WriteByte(hPort, (dwData >> 24) & 0xFF);
    WriteByte(hPort, (dwData >> 16) & 0xFF);
    WriteByte(hPort, (dwData >> 8) & 0xFF);
    WriteByte(hPort, dwData & 0xFF);
}

/* Write a block of wBytesToWrite from pbBuffer */
void WriteBlock(HPORT hPort, PBYTE pbBuffer, WORD wBytesToWrite)
{
    BYTE bByte;

    /* until all bytes are sent */
    while (wBytesToWrite > 0)
    {
        /* get next byte */
        bByte = *pbBuffer;
        /* send it out */
        WriteByte(hPort, bByte);
        /* prepare for next one */
        pbBuffer++;
        wBytesToWrite--;
    }
}

/* Truncate filename to 8.3 notation; returns TRUE if truncated */
BOOL TruncFilename8Dot3(PCSTR pcszInput, PBYTE pbOutput, WORD wLength)
{
    int iPos = 0;
    int iCount = 0;
    BOOL fTruncated = FALSE;

    /* Get name part */
    while ((iCount < 8) && (*pcszInput != '\0') && (*pcszInput != '.'))
    {
        pbOutput[iPos++] = *pcszInput;
        pcszInput++;
        iCount++;
    }

    /* String already finished, fill up with zero and return */
    if (*pcszInput == '\0')
    {
        while (iPos < wLength)
            pbOutput[iPos++] = '\0';

        return fTruncated;
    }

    /* If there was no dot truncation took place */
    if (*pcszInput != '.')
    {
        pcszInput++;
        fTruncated = TRUE;
    }

    /* Go forward until dot is found, if any */
    while ((*pcszInput != '\0') && (*pcszInput != '.'))
        pcszInput++;

    /* Dot following? */
    if (*pcszInput == '.')
    {
        /* Insert dot */
        pbOutput[iPos++] = '.';
        pcszInput++;
    }

    /* Get extension part (first extension if more than one exist) */
    iCount = 0;
    while ((iCount < 3) && (*pcszInput != '\0') && (*pcszInput != '.'))
    {
        pbOutput[iPos++] = *pcszInput;
        pcszInput++;
        iCount++;
    }

    /* If next character is not '\0' truncation did happen */
    fTruncated = fTruncated || (*pcszInput != '\0');

    /* Fill up with zero and return */
    while (iPos < wLength)
        pbOutput[iPos++] = '\0';

    return fTruncated;
}

/* Generates a simple checksum by adding all words together */
WORD GenerateChecksum(PBYTE pbDataBlock, WORD wBlockLen)
{
    WORD wChecksum = 0;

    /* while whole words are available */
    while (wBlockLen > 1)
    {
        wChecksum += *((PWORD)pbDataBlock);
        pbDataBlock += 2;
        wBlockLen -= 2;
    }

    /* odd bytes are padded with zero on the right */
    if (wBlockLen > 0)
    {
        wChecksum += (*pbDataBlock) << 8;
    }

    return wChecksum;
}

/* Print out statistic information and a progress bar */
void PrintProgress(PCSTR pcszFileName, DWORD dwTotal,
                   DWORD dwCurrent, BOOL fError)
{
    int iFilledSteps;
    int iStepCount;

    /* steps in progress bar that are filled */
    iFilledSteps = dwCurrent * BAR_WIDTH / dwTotal;

    /* draw progress bar */
    for (iStepCount = 0; iStepCount < BAR_WIDTH; iStepCount++)
    {
        putchar((iStepCount < iFilledSteps) ?
                BAR_FILLED_CHAR : BAR_HOLLOW_CHAR);
    }

    /* add some textual information */
    printf(" %12s %lu/%lu ", pcszFileName, dwCurrent, dwTotal);

    if (fError)
    {
        /* in error condition */
        printf("(ERROR)");
    }
    else
    {
        /* print the progress in percent */
        printf("(%.1f%%)", dwCurrent * 100.0 / dwTotal);
    }

    /* overwrite this line next time */
    putchar('\r');
    fflush(stdout);
}

/* Receives a file and saves it to pcszFileName */
int RecvFile(HPORT hPort)
{
    FILE *pFile;
    DWORD dwTotalBytesToRead;
    DWORD dwTotalBytesRead = 0;
    BYTE abCurrentBlock[BLOCK_SIZE];
    WORD wBytesToRead;
    WORD wSenderChecksum;
    WORD wLocalChecksum;
    BYTE abFileName[MAX_FILENAME_LEN + 1];
    BYTE bTransRequest;

    do
    {
        bTransRequest = ReadByte(hPort);
    }
    while (bTransRequest != TRANS_REQUEST);
    WriteByte(hPort, TRANS_ACCEPTED);

    /* Get name of file to receive */
    ReadBlock(hPort, abFileName, MAX_FILENAME_LEN);
    /* Zero-terminate it */
    abFileName[MAX_FILENAME_LEN] = '\0';

    /* Get number of bytes to receive */
    dwTotalBytesToRead = ReadDword(hPort);

    /* Open file for output */
    pFile = fopen(abFileName, "wb");
    if (pFile == NULL)
    {
        fprintf(stderr, "Couldn't open %s for writing: %s",
                abFileName, strerror(errno));
        /* Notify sender that this file is declined */
        WriteByte(hPort, FILE_DECLINED);
        return EXIT_FAILURE;
    }

    PrintProgress(abFileName, dwTotalBytesToRead, 0, FALSE);

    /* Notify sender that this file is accepted */
    WriteByte(hPort, FILE_ACCEPTED);

    /* While not finished with reading */
    while (dwTotalBytesRead < dwTotalBytesToRead)
    {
        /* Read a whole block unless the remaining rest is smaller */
        if (dwTotalBytesToRead - dwTotalBytesRead < BLOCK_SIZE)
            wBytesToRead = dwTotalBytesToRead - dwTotalBytesRead;
        else
            wBytesToRead = BLOCK_SIZE;

        /* Read the block */
        ReadBlock(hPort, abCurrentBlock, wBytesToRead);

        /* Read the sender's checksum */
        wSenderChecksum = ReadWord(hPort);

        /* Calculate the checksum over the received data */
        wLocalChecksum = GenerateChecksum(abCurrentBlock, wBytesToRead);

        /* If the checksums match... */
        if (wSenderChecksum == wLocalChecksum)
        {
            /* ...save data to file */
            fwrite(abCurrentBlock, 1, wBytesToRead, pFile);
            dwTotalBytesRead += wBytesToRead;
        }

        /* Update status */
        PrintProgress(abFileName,
                      dwTotalBytesToRead,
                      dwTotalBytesRead,
                      wSenderChecksum != wLocalChecksum);

        /* Tell sender the result of checksum comparing; the sender will
         * repeat the block if the checksums differ.
         */
        WriteWord(hPort, wSenderChecksum + ~wLocalChecksum);
    }

    /* Done with this file, next line */
    putchar('\n');

    /* Close the file to make sure things get written through */
    fclose(pFile);
    return EXIT_SUCCESS;
}

/* Send the file pcszFileName */
int SendFile(HPORT hPort, PCSTR pcszFileName)
{
    FILE *pFile;
    DWORD dwTotalBytesToWrite;
    DWORD dwTotalBytesWritten = 0;
    BYTE abCurrentBlock[BLOCK_SIZE];
    WORD wBytesToWrite;
    WORD wChecksumResponse = CHECKSUM_OK;
    WORD wLocalChecksum;
    BYTE bAcceptanceResponse;
    struct stat stbuf;
    BYTE abFileName8Dot3[MAX_FILENAME_LEN];

    WriteByte(hPort, TRANS_REQUEST);
    ReadByte(hPort);

    /* Send filename (as 8.3) */
    TruncFilename8Dot3(pcszFileName, abFileName8Dot3, MAX_FILENAME_LEN);
    WriteBlock(hPort, abFileName8Dot3, MAX_FILENAME_LEN);

    /* Get number of bytes to send */
    if (stat((PSTR)pcszFileName, &stbuf) < 0)
    {
        fprintf(stderr, "Couldn't stat %s: %s\n",
                pcszFileName, strerror(errno));
        return EXIT_FAILURE;
    }

    dwTotalBytesToWrite = stbuf.st_size;

    /* Open file for input */
    pFile = fopen((PSTR)pcszFileName, "rb");
    if (pFile == NULL)
    {
        fprintf(stderr, "Couldn't open %s for reading: %s",
                pcszFileName, strerror(errno));
        return EXIT_FAILURE;
    }

    /* Send number of bytes */
    WriteDword(hPort, dwTotalBytesToWrite);

    /* Wait for response */
    bAcceptanceResponse = ReadByte(hPort);

    if (bAcceptanceResponse != FILE_ACCEPTED)
    {
        printf("file %s declined by receiver\n", pcszFileName);
        return EXIT_FAILURE;
    }

    PrintProgress(pcszFileName, dwTotalBytesToWrite, 0, FALSE);

    /* While not finished with writing */
    while (dwTotalBytesWritten < dwTotalBytesToWrite)
    {
        /* Write a whole block unless the remaining rest is smaller */
        if (dwTotalBytesToWrite - dwTotalBytesWritten < BLOCK_SIZE)
            wBytesToWrite = dwTotalBytesToWrite - dwTotalBytesWritten;
        else
            wBytesToWrite = BLOCK_SIZE;

        /* Read the block from file if last block was OK */
        if (wChecksumResponse == CHECKSUM_OK)
        {
            fread(abCurrentBlock, 1, wBytesToWrite, pFile);
        }

        /* Write the block */
        WriteBlock(hPort, abCurrentBlock, wBytesToWrite);

        /* Calculate the checksum over the sent data */
        wLocalChecksum = GenerateChecksum(abCurrentBlock, wBytesToWrite);

        /* Send the checksum for this block */
        WriteWord(hPort, wLocalChecksum);

        /* Read the response from the receiver */
        wChecksumResponse = ReadWord(hPort);

        /* Was the checksum OK? */
        if (wChecksumResponse == CHECKSUM_OK)
        {
            dwTotalBytesWritten += wBytesToWrite;
        }

        /* Update status */
        PrintProgress(pcszFileName, dwTotalBytesToWrite, dwTotalBytesWritten,
                      wChecksumResponse != CHECKSUM_OK);
    }

    /* Done with this file, next line */
    putchar('\n');

    /* Close the file */
    fclose(pFile);
    return EXIT_SUCCESS;
}

int main(int argc, char *argv[])
{
    int iIndex;
    int iFileCount = 0;
    HPORT hPort;
    PCSTR *ppcszFileNames = NULL;
    PCSTR pcszPortName = DEF_COMPORT;
    DWORD dwBaudRate = DEF_BAUDRATE;
    PCSTR pcszModeString = DEF_MODESTRING;

    /* Get enough memory to save all filenames that may come */
    ppcszFileNames = calloc(argc, sizeof(PCSTR));

    if (ppcszFileNames == NULL)
    {
        fprintf(stderr, "Out of memory, try fewer files\n");
        return EXIT_FAILURE;
    }

    for (iIndex = 1; iIndex < argc; iIndex++)
    {
        /* print help and exit */
        if (!strcmp(argv[iIndex], "-h") ||
            !strcmp(argv[iIndex], "-?"))
        {
            PrintHelp();
            return EXIT_SUCCESS;
        }

        /* get port to send on */
        if (!strcmp(argv[iIndex], "-p"))
        {
            pcszPortName = NULL;
            if (iIndex + 1 < argc)
            {
                pcszPortName = argv[++iIndex];
                continue;
            }
        }

        /* get baudrate to send with */
        if (!strcmp(argv[iIndex], "-b"))
        {
            dwBaudRate = 0;
            if (iIndex + 1 < argc)
            {
                dwBaudRate = strtoul(argv[++iIndex], NULL, 0);
                continue;
            }
        }

        /* get mode to use */
        if (!strcmp(argv[iIndex], "-m"))
        {
            pcszModeString = NULL;
            if (iIndex + 1 < argc)
            {
                pcszModeString = argv[++iIndex];
                continue;
            }
        }

        /* get file to send */
        ppcszFileNames[iFileCount++] = argv[iIndex];
    }

    /* Port name specified? */
    if (pcszPortName == NULL)
    {
        fprintf(stderr, "missing port name\n");
        return EXIT_FAILURE;
    }

    /* Baudrate specified? */
    if (dwBaudRate == 0)
    {
        fprintf(stderr, "missing baudrate\n");
        return EXIT_FAILURE;
    }

    /* Mode string specified? */
    if (pcszModeString == NULL)
    {
        fprintf(stderr, "missing mode string\n");
        return EXIT_FAILURE;
    }

    /* Open COM port */
    hPort = OpenComPort(pcszPortName);
    if (hPort == INVALID_PORT_HANDLE)
    {
        fprintf(stderr, "could not open %s\n", pcszPortName);
        return EXIT_FAILURE;
    }

    /* Configure COM port */
    if (ConfigureComPort(hPort, dwBaudRate, pcszModeString) == FALSE)
    {
        fprintf(stderr, "couldn't configure port\n");
        return EXIT_FAILURE;
    }

    if (iFileCount > 0)
    {
        for (iIndex = 0; iIndex < iFileCount; iIndex++)
        {
            SendFile(hPort, ppcszFileNames[iIndex]);
        }
    }
    else
    {
        while (RecvFile(hPort) != EXIT_FAILURE)
            ;
    }

    /* Close COM port */
    CloseComPort(hPort);

    if (ppcszFileNames != NULL)
        free(ppcszFileNames);

    return EXIT_SUCCESS;
}
